//alle Listings zu C++-Tricks 2: Fold Expressions


// Listing 1: Funktion _StrCat_ zum Verknüpfen mehrerer Strings mit C++11-Mitteln

#include <iostream>
#include <string>

namespace details {
    void StrCat(std::string& buffer) {}

    template<typename T, typename... Ts>
    void StrCat(std::string& buffer, T&& t, Ts&&... ts)
    {
        buffer += t;

        StrCat(buffer, ts...);
    }
}  // namespace details

template<typename... Ts>
std::string StrCat(Ts&&... ts)
{
    std::string ret{};

    details::StrCat(ret, ts...);

    return ret;
}

int main()
{
    std::string a{"Hello,"};
    std::string b{" World"};
    std::string c{"!"};

    auto x = StrCat(a, b, c);

    std::cout << x;
}

//Ende Listing 1



// Listing 2: Die Funktion _StrCat_ verbessert mit der Funktion _TotalLen_

#include <cstdio>
#include <iostream>

namespace details {

    size_t TotalLen() { return 0; }

    template<typename T, typename... Ts>
    size_t TotalLen(T&& t, Ts&&... ts)
    {
        return t.size() + TotalLen(ts...);
    }

    void StrCat(std::string& buffer) {}

    template<typename T, typename... Ts>
    void StrCat(std::string& buffer, T&& t, Ts&&... ts)
    {
        buffer += t;

        StrCat(buffer, ts...);
    }
}  // namespace details

template<typename... Ts>
std::string StrCat(Ts&&... ts)
{
    std::string ret{};

    ret.reserve(details::TotalLen(ts...));

    details::StrCat(ret, ts...);

    return ret;
}

int main()
{
    std::string a{"Hello,"};
    std::string b{" World"};
    std::string c{"!"};

    auto x = StrCat(a, b, c);

    std::cout << x;
}

//Ende Listing 2



// Listing 3: Die Funktion _StrCat_ mit Fold Expressions entrümpelt

#include <cstdio>
#include <iostream>

template<typename... Ts>
auto StrCat(Ts&&... ts)
{
    std::string ret{};

    ret.reserve((ts.size() + ...));

    (ret += ... += ts);

    return ret;
}

int main()
{
    std::string a{"Hello,"};
    std::string b{" World"};
    std::string c{"!"};

    auto x = StrCat(a, b, c);

    std::cout << x << '\n';
}

//Ende Listing 3



// Listing 4: Programm mit wiederholten Aufrufen von _v.push_back_

#include <cstdio>
#include <vector>

void ManipulateVector(std::vector<int>& v)
{
    // manipulate v
}

int main()
{
    std::vector<int> v{1, 2, 3};

    ManipulateVector(v);

    int z = 5;

    v.push_back(z);
    v.push_back(6);
    v.push_back(7);
}

//Ende Listing 4



// Listing 5: Mit Fold Expressions entfallen die wiederholten Aufrufe

#include <cstdio>
#include <vector>

void ManipulateVector(std::vector<int>& v)
{
    // manipulate v
}

template<typename T, typename... Args>
void PushBack(std::vector<T>& v, Args&&... args)
{
    (v.push_back(std::forward<Args>(args)), ...);
}

int main()
{
    std::vector<int> v{1, 2, 3};

    int z = 5;

    ManipulateVector(v);

    PushBack(v, z, 6, 7);
}

//Ende Listing 5



// Listing 6: Mit Templates und Hilfsfunktionen ein variadisches Template bauen

template<typename T, size_t N>
constexpr bool Compare(const T (&a)[N], const T (&b)[N])
{
    return details::Compare(a, b, std::make_index_sequence<N>{});
}

//Ende Listing 6



// Listing 7: details::Compare als variadische Funktion

namespace details {
    template<typename T, size_t N, size_t... I>
    constexpr bool
    Compare(const T (&a)[N], const T (&b)[N], std::index_sequence<I...>)
    {
        return ((a[I] == b[I]) && ...);
    }
}  // namespace details

//Ende Listing 7



// Listing 8: Die Funktion compare in einer C++20-Implementierung mit Templated Lambdas 

template<typename T, size_t N>
constexpr bool Compare(const T (&a)[N], const T (&b)[N])
{
    return [&]<size_t... I>(std::index_sequence<I...>)
    {
        return ((a[I] == b[I]) && ...);
    }
    (std::make_index_sequence<N>{});
}

//Ende Listing 8



// Listing 9: Die variadische _compare_-Funktion mit festem Vergleichswert

namespace details {
    template<typename T, size_t N, size_t... I>
    constexpr bool
    Compare(const T (&a)[N], const T& b, std::index_sequence<I...>)
    {
        return ((a[I] == b) && ...);
    }
}  // namespace details

template<typename T, size_t N>
constexpr bool Compare(const T (&a)[N], const T& b)
{
    return details::Compare(a, b, std::make_index_sequence<N>{});
}

//Ende Listing 9



// Listing 10: variadische Funktion in_range

template<typename T, typename... Ts>
constexpr bool in_range(T min, T max, Ts... ts)
{
    return ((min <= ts && ts <= max) && ...);
}

//Ende Listing 10



// Listing 11: Hilfs-Type-Traits für die in_range-Konstruktion

/* Abschnitt 1: Remove-Type-Traits */

template<typename T>
struct remove_cvref
{
    using type = std::remove_cv_t<std::remove_reference_t<T>>;
};

template<typename T>
using remove_cvref_t = typename remove_cvref<T>::type;

/* Abschnitt 2: Parameterprüfung */

template<typename T, typename... Ts>
struct are_same
{
    static constexpr bool value = std::conjunction_v<
        std::is_same<remove_cvref_t<T>, remove_cvref_t<Ts>>...>;
};
template<typename T, typename... Ts>
constexpr bool are_same_v = are_same<T, Ts...>::value;

/* Abschnitt 3: Bestimmung des ersten Typs */

template<typename T, typename...>
struct first_type
{
    using type = remove_cvref_t<T>;
};

/* Abschnitt 4: is-Objekt */

template<typename... Ts>
struct is
{
    /* Abschnitt 5: Basistyp definieren */

    using BaseType = typename first_type<Ts...>::type;

    /* Abschnitt 6: Typ-Prüfung und Fehlerausgabe */

    static_assert(are_same_v<BaseType, Ts...>,
                  "Only parameters of the same type are allowed.");

    /* Abschnitt 7: Größe des Parameter-Pack */

    static constexpr auto SIZE = sizeof...(Ts);
    const BaseType        values[SIZE];

    /* Abschnitt 8: Array values initialisieren */

    constexpr is(Ts&&... ts)
    : values{std::forward<Ts>(ts)...}
    {
    }

    /* Abschnitt 9: in_range initialisieren */

    bool in_range(const BaseType& min, const BaseType& max) const;

    /* Abschnitt 10: make_index_sequence-Trick */

    constexpr bool in_range(const BaseType& min, const BaseType& max) const
    {
        return compare(min, max, std::make_index_sequence<SIZE>{});
    }

    /* Abschnitt 11: Hilftsmethode compare */

    template<std::size_t... I>
    constexpr bool compare(const BaseType& min,
                           const BaseType& max,
                           std::index_sequence<I...>) const
    {
        return (((min <= values[I]) && (values[I] <= max)) && ...);
    }
};

/* Abschnitt 12: Ein is-Objekt ohne die spezifische Angabe von Parametern erstellen */

template<typename... Ts>
is(Ts&&...)->is<Ts&&...>;  // CTAD

//Ende Listing 11



// Listing 12: Gleichheitsvergleich für ìs

template<std::size_t... I>
constexpr bool compare_equal(const BaseType& value, std::index_sequence<I...>) const
{
    return ((value == values[I]) && ...);
}

constexpr bool equal(const BaseType& value) const
{
    return compare_equal(value, std::make_index_sequence<SIZE>{});
}

constexpr bool operator==(const BaseType& value) const
{
    return equal(value);
}

friend constexpr bool operator==(const BaseType& lhs, const is& rhs)
{
    return rhs == lhs;
}

//Ende Listing 12
